@@ -6,16 +6,7 @@
 #
 #     https://github.com/ASukhanov/apstrim/blob/main/LICENSE
 #
-#__version__ = '1.0.3 2021-06-01'# EPICS and LITE support is OK, Compression supported
-#__version__ = '1.0.4a 2021-06-11'# flush the file after each section
-#__version__ = '1.0.4 2021-06-11'# if file exist then rename the existing file
-#__version__ = '1.0.5 2021-06-14'# handling of different returned maps
-#__version__ = '1.0.6 2021-06-19'# filename moved from instantiation to new method: start(), timestamp is int(nanoseconds)
-#__version__ = '1.0.7 2021-06-20'# Docstrings updated
-#__version__ = '1.0.9 2021-06-21'# new keyword: self.use_single_float
-#__version__ = '1.0.10 2021-06-22'# separate events for exit and stop
-#__version__ = '1.0.11 2021-06-22'# intercept exception in _delivered()
-__version__ = '1.1.0 2021-06-23'# fixed bug when subscriptions was multiplied every start()
+__version__ = '1.2.0 2021-07-19'# random-access-ready, bug __printe fixed
 
 #TODO: consider to replace msgpack_numpy with something simple and predictable.
 #The use_single_float has no efect in msgPack,
@@ -34,13 +25,15 @@ import msgpack_numpy
 msgpack_numpy.patch()
 
 #````````````````````````````Globals``````````````````````````````````````````
-SecDateTime, SecParagraph = 0,1
-Nano = 1000000000
+Nano = 0.000000001
 #````````````````````````````Helper functions`````````````````````````````````
 def _printTime(): return time.strftime("%m%d:%H%M%S")
 def _printi(msg): print(f'INFO_AS@{_printTime()}: {msg}')
 def _printw(msg): print(f'WARN_AS@{_printTime()}: {msg}')
 def _printe(msg): print(f'ERROR_AS@{_printTime()}: {msg}')
+def _printd(msg):
+    if apstrim.Verbosity>0:
+        print(f'DBG_AS: {msg}')
 
 def croppedText(txt, limit=200):
     if len(txt) > limit:
@@ -57,60 +50,80 @@ def shortkey(i:int):
 #````````````````````````````Serializer class`````````````````````````````````
 class apstrim():
     """Create the object streamer. 
-    publisher:      is a class, providing a subscribe() method,
-    devPar:         list of device:parameter strings,
+    publisher:  is a class, providing a subscribe() method,
+    devPar:     list of device:parameter strings,
     sectionInterval: time between writing of the logBook sections
     compression:    compression enable flag,
-    quiet:          do not print section writing progress,
-    use_single_float: Use single precision float type for float. (default: False)
-    eventExit:      is a threading.Event, which will be set when application
-     is about to exit.
+    quiet:      do not print section writing progress,
+    use_single_float: Use single precision float type for float. (default: True)
+    dirSize:    size of the table of contents, which is usd for random-access
+                retrieval. If 0 then no table will be created.
+        Class properties:
+    eventExit:  Is a threading.Event, which will be set for safe exit.
+    Verbosity:  Show more log messages
     """
     eventExit = threading.Event()
     _eventStop = threading.Event()
+    Verbosity = 0
 
     def __init__(self, publisher, devPars:list, sectionInterval=60.
-    , compress=False, quiet=False, use_single_float=True):
+    , compress=False, quiet=False, use_single_float=True, dirSize=0):
         #_printi(f'apstrim  {__version__}, sectionInterval {sectionInterval}')
+        signal.signal(signal.SIGINT, _safeExit)
+        signal.signal(signal.SIGTERM, _safeExit)
+
+        self.lock = threading.Lock()
         self.publisher = publisher
         self.devPars = devPars
         self.sectionInterval = sectionInterval
         self.quiet = quiet
         self.use_single_float = use_single_float
-        signal.signal(signal.SIGINT, _safeExit)
-        signal.signal(signal.SIGTERM, _safeExit)
 
-        # create section Abstract
-        self.abstractSection = {'apstrim ':__version__
-        , 'sectionInterval':sectionInterval}
+        # table of contents - related variables
+        self.dirSize = dirSize
+        self.contents_downsampling_factor = 1# No downsampling 
+
+        # create a section Abstract
+        self.abstractSection = {'abstract':{'apstrim ':__version__
+        , 'sectionInterval':sectionInterval}}
+        abstract = self.abstractSection['abstract']
+
         if compress:
             import lz4framed
             self.compress = lz4framed.compress
-            self.abstractSection['compression'] = 'lz4framed'
+            abstract['compression'] = 'lz4framed'
         else:
             self.compress = None
-            self.abstractSection['compression'] = 'None'
+            abstract['compression'] = 'None'
         _printi(f'Abstract section: {self.abstractSection}')
 
-        # create section Abbreviations
+        # a section has to be created before subscription
+        self._create_logSection()
+
+        # subscribe to parameters
         self.pars = {}
-        for i,pname in enumerate(self.devPars):
+        #for i,pname in enumerate(self.devPars):
+        i = 0
+        for pname in self.devPars:
             devPar = tuple(pname.rsplit(':',1))
-            if True:#try:
+            try:
                 self.publisher.subscribe(self._delivered, devPar)
-            else:#except Exception as e:
-                __printe(f'Subscription failed for {pname}: {e}')
+            except:# Exception as e:
+                _printe(f'Could not subscribe  for {pname}')#: {e}')
                 continue
             self.pars[pname] = [shortkey(i)]
+            i += 1
+        if len(self.pars) == 0:
+            _printe(f'Could not build the list of parameters')
+            sys,exit()
         _printi(f'parameters: {self.pars}')
-        self.abbreviationSection = msgpack.packb({'parameters':self.pars}
-        , use_single_float=self.use_single_float)
 
-        self.lock = threading.Lock()
+        self.abbreviationSection = msgpack.packb({'abbreviations':self.pars}
+        , use_single_float=self.use_single_float)
 
     def start(self, fileName='apstrim.aps'):
         """Start streaming of the data objects to logbook file.
-        If file is already exist then it will be renamed and
+        If file is already exists then it will be renamed and
         a new file will be open with the provided name.
         """
         self._eventStop.clear()
@@ -128,15 +141,25 @@ class apstrim():
 
         self.logbook = open(fileName, 'wb')
 
+        # write a preliminary 'Table of contents' section
+        if self.dirSize:
+            self.contentsSection = {'contents':{'size':self.dirSize}, 'data':{}}
+            self.dataContents = self.contentsSection['data']
+            self.logbook.write(msgpack.packb(self.contentsSection))
+            # skip the 'Table of contents' zone of the logbook
+            self.logbook.seek(self.dirSize)
+
         # write the sections Abstract and Abbreviations
+        _printd(f'write Abstract@{self.logbook.tell()}')
         self.logbook.write(msgpack.packb(self.abstractSection
         , use_single_float=self.use_single_float))
+        _printd(f'write Abbreviations@{self.logbook.tell()}')
         self.logbook.write(self.abbreviationSection)
 
         self._create_logSection()
 
         #_printi('starting serialization  thread')
-        myThread = threading.Thread(target=self._serialize_section)
+        myThread = threading.Thread(target=self._serialize_sections)
         myThread.start()
 
         _printi(f'Logbook file: {fileName} created')
@@ -151,7 +174,7 @@ class apstrim():
         Called when the requested data have been changed.
         args is a map of delivered objects."""
         #print(f'delivered: {args}')
-        timestampedMap = {}
+        #self.timestampedMap = {}
         for devPar,props in args[0].items():
             #print(f'devPar: {devPar,props}, {type(devPar)}')
             try:
@@ -161,10 +184,10 @@ class apstrim():
                 value = props['value']
                 timestamp = props.get('timestamp')# valid in EPICS and LITE
                 if timestamp == None:# decode ADO timestamp 
-                    timestamp = int(props['timestampSeconds']*Nano
+                    timestamp = int(props['timestampSeconds']/Nano
                     + props['timestampNanoSeconds'])
                 else:
-                    timestamp = int(timestamp*Nano)
+                    timestamp = int(timestamp/Nano)
                 skey = self.pars[dev+':'+par][0]
               elif devPar == 'ppmuser':# ADO has extra item, skip it.
                 continue
@@ -177,7 +200,7 @@ class apstrim():
                         timestamp = int(pars[par]['t'])
                     except: # try old LITE packing
                         value = pars[par]['value']                     
-                        timestamp = int(pars[par]['timestamp']*Nano)
+                        timestamp = int(pars[par]['timestamp']/Nano)
                     skey = self.pars[devPar+':'+par][0]
             except Exception as e:
                 _printw(f'exception in unpacking: {e}')
@@ -189,62 +212,107 @@ class apstrim():
                 try:    
                     if value.dtype=='float64':
                         value = value.astype('float32')
-                    #print(f'Numpy f64->f32 reduction time: {round(timer()-ts,6)}')
+                    #print(f'Numpy f64->f32 downsampling time: {round(timer()-ts,6)}')
                 except:    pass
-
-            if timestamp in timestampedMap:
-                timestampedMap[timestamp][skey] = value
+            #print(f'ts:{timestamp}, keys:{self.timestampedMap.keys()}')
+            #print(f'ts:{timestamp}, tsMap:{self.timestampedMap}')
+            if timestamp in self.timestampedMap:
+                #print(f'add to self.timestampedMap: {timestamp,skey}')
+                self.timestampedMap[timestamp][skey] = value
             else:
-                timestampedMap[timestamp] = {skey:value}
-        #TODO: timestampedMap may need sorting
-        #print(f'timestampedMap: {timestampedMap}')
-        with self.lock:
-            try:
-                self.logParagraph.append(list(timestampedMap.items())[0])
-            except Exception as e:
-                _printe(f'Exception in _delivered for {args}: {e}')
+                #print(f'    create entry in self.timestampedMap: {timestamp,skey}')
+                self.timestampedMap[timestamp] = {skey:value}
+            #print(f'devPar {devPar}@{timestamp}, tsMap: {self.timestampedMap[timestamp].keys()}')
+        #TODO: self.timestampedMap may need sorting
+        #print(f'self.timestampedMap: {self.timestampedMap}')
         
     def _create_logSection(self):
       with self.lock:
-        self.logParagraph = []
-        key = time.strftime("%y%m%d:%H%M%S")
-        self.logSection = (key, self.logParagraph)
+        #print('create empty list of paragraphs')
+        self.sectionKey = time.strftime("%y%m%d_%H%M%S")
+        self.timestampedMap = {}
 
-    def _serialize_section(self):
-        #_printi('serialize_section started')
+    def _serialize_sections(self):
+        #_printi('serialize_sections started')
         periodic_update = time.time()
-        stat = [0, 0]
+        statistics = [0, 0, 0, 0]#
+        NSections, NParagraphs, BytesRaw, BytesFinal = 0,1,2,3
         try:
           while not self._eventStop.is_set():
             self._eventStop.wait(self.sectionInterval)
-            if len(self.logSection[SecParagraph]) == 0:
+            if len(self.timestampedMap) == 0:
                 continue
 
-            #print(f'write section {self.logSection}')
+            # register the section in the table of contents, this should be skipped
+            # for some entries when the contents downsampling is active.
+            if self.dirSize:
+                #offset = statistics[BytesRaw] + self.dirSize
+                #self.dataContents[self.sectionKey] = offset
+                rf = self.contents_downsampling_factor
+                if rf <=1 or (statistics[NSections]%rf) == 0:
+                    self.dataContents[self.sectionKey] = self.logbook.tell()
+                    packed = msgpack.packb(self.contentsSection)
+                    if len(packed) < self.dirSize:
+                        self.packedContents = packed
+                    else:
+                        _printw((f'The contents size is too small for'
+                        f' {len(packed)} bytes. Half of the entries will be'
+                        ' removed to allow for more entries.}'))
+                        self.contents_downsampling_factor *= 2
+                        downsampled_contents = dict(list(self.dataContents.items())\
+                          [::self.contents_downsampling_factor])
+                        self.contentsSection['data'] = downsampled_contents
+                        _printd(f'downsampled contentsSection:{self.contentsSection}')
+                        self.packedContents = msgpack.packb(self.contentsSection)
+
+            #print(f'section is ready, write it to logbook {self.logSection}')
+            _printd(f'writing section {statistics[NSections]} @ {self.logbook.tell()}')
+            statistics[NSections] += 1
+            with self.lock:
+                paragraphs = list(self.timestampedMap.items())
+                self.logSection = (self.sectionKey, paragraphs)
+                #print(f'section:{self.logSection}')
             packed = msgpack.packb(self.logSection
             , use_single_float=self.use_single_float)
+            statistics[NParagraphs] += len(paragraphs)
+            statistics[BytesRaw] += len(packed)
             if self.compress is not None:
                 compressed = self.compress(packed)
                 packed = msgpack.packb(compressed
                 , use_single_float=self.use_single_float)
+            statistics[BytesFinal] += len(packed)
             self.logbook.write(packed)
             self.logbook.flush()
 
-            stat[0] += len(self.logSection[SecParagraph])
-            stat[1] += len(packed)
             self._create_logSection()
             timestamp = time.time()
             dt = timestamp - periodic_update
             if dt > 10.:
                 periodic_update = timestamp
                 if not self.quiet:
-                    print(f'{time.strftime("%y-%m-%d %H:%M:%S")} Logged {stat[0]} paragraphs, {stat[1]/1000.} KBytes')
+                    print(f'{time.strftime("%y-%m-%d %H:%M:%S")} Logged'
+                    f' {statistics[NSections]} sections,'
+                    f' {statistics[NParagraphs]} paragraphs,'
+                    f' {statistics[BytesFinal]/1000.} KBytes')
         except Exception as e:
-            print(f'ERROR: Exception in serialize_section: {e}')
-        print(f'Logging finished after {stat[0]} paragraphs')
+            print(f'ERROR: Exception in serialize_sections: {e}')
+
+        # logging finished
+        # rewrite the contentsSection
+        if self.dirSize:
+            self.logbook.seek(0)
+            self.logbook.write(self.packedContents)
+
+        # print status
+        msg = (f'Logging finished for {statistics[NSections]} sections,'
+        f' {statistics[NParagraphs]} paragraphs,'
+        f' {statistics[BytesFinal]/1000.} KB.')
+        if self.compress is not None:
+            msg += f' Compression ratio:{round(statistics[BytesRaw]/statistics[BytesFinal],2)}'
+        print(msg)
         self.logbook.close()
                 
-def _safeExit(_signo, _stack_frame):#, self=None):
+def _safeExit(_signo, _stack_frame):
     print('safeExit')
     apstrim._eventStop.set()
     apstrim.eventExit.set()
